var async = require('async')
  , cli = require('../lib')
  , fs = require('fs')
  , moment = require('moment')
  , nitrogen = require('nitrogen');

var help = function(callback) {

    cli.utils.displayCommands('Message Commands', [
        ['ls [filter]', 'Show messages matching an optional filter. (eg. ls \'{"type": "image"}\''],
        ['import <file>', 'Imports a message stream from <file>'],
        ['replay <file> --loop --maxGap <20>', 'Replays a message stream from <file>'],
        ['send [message]', 'Send a message.  (eg. send \'{"type": "_temp", "body": { "temp": 15.5 } }\'']
    ]);

    return callback();
};

var importFile = function(callback) {
    var filePath = cli.nextArgument();
    fs.readFile(filePath, function(err, jsonString) {
        if (err) return callback(err);

        var messages = JSON.parse(jsonString);

        cli.globals.startSession(function(err, session, user) {
            if (err) return callback(err);

            var start = 0;

            console.dir(messages);
            console.log(start);
            console.log(messages.length);

            async.whilst(
                function() { return start < messages.length },
                function(batchCallback) {
                    var batch = messages.slice(start, 1)
                                        .map(function(batchObject) {
                                            return new nitrogen.Message(batchObject);
                                        });
                    start += 1;
                    console.dir(batch);
                    nitrogen.Message.sendMany(session, batch, batchCallback);
                },
                callback
            );
        });
    });
};

var list = function(callback) {
    var filterString = cli.nextArgument();
    var filter;

    if (filterString) {
        filter = JSON.parse(filterString);
    } else {
        filter = {};
    }

    cli.globals.startSession(function(err, session, user) {
        if (err) return callback(err);

        // TODO: Add in limit, sorts, skip flags.
        var options = { limit: 10, sort: { ts: -1 } };

        nitrogen.Message.find(session, filter, options, function(err, messages) {
            if (err) return callback(err);

            var widths = [23, 25, 25, 15, 17];
            var columns = ['TIMESTAMP', 'FROM', 'TO', 'TYPE', 'BODY'];

            cli.log.info(cli.utils.prettyPrint(columns, widths));

            messages.forEach(function(message) {
                var columns = [
                    moment(message.ts).format('MM/DD/YYYY HH:mm:ss.SS'),
                    message.from,
                    message.to || "",
                    message.type,
                    JSON.stringify(message.body || "")
                ];

                cli.log.info(cli.utils.prettyPrint(columns, widths));
            });

            return callback();
        });
    });
};

var replay = function(callback) {
    var filePath = cli.nextArgument();
    var maxGap = cli.flags.maxGap || 25;

    fs.readFile(filePath, function(err, jsonString) {
        if (err) return callback(err);

        var messages = JSON.parse(jsonString);

        var lastTimestamp;

        cli.globals.startSession(function(err, session, user) {
            if (err) return callback(err);

            var batchInterval = 2500;
            var currentIndex = 0;
            var startTimestamp = Date.parse(messages[0].ts);
            var deltaToOldTimestamp = new Date().getTime() - startTimestamp;
            var endTimestamp = Date.parse(messages[0].ts);

            setInterval(function() {

                if (currentIndex >= messages.length)
                    process.exit(0);

                endTimestamp = startTimestamp + batchInterval;

                var batch = [];
                var shouldContinue = true;

                while (shouldContinue) {
                    var message = new nitrogen.Message(messages[currentIndex]);
                    message.ts = new Date(Date.parse(message.ts) + deltaToOldTimestamp);

                    batch.push(message);

                    currentIndex++;

                    if (currentIndex < messages.length) {
                        shouldContinue = Date.parse(messages[currentIndex].ts) < endTimestamp;
                    } else {
                        shouldContinue = false;
                        if (cli.flags.loop) {
                            currentIndex = 0;
                            endTimestamp = Date.parse(messages[0].ts);
                        }
                    }
                }

                if (currentIndex < messages.length)
                    startTimestamp = Math.max(endTimestamp, Date.parse(messages[currentIndex].ts) - maxGap * 60 * 1000);

                if (batch.length === 0) {
                    var delta = Date.parse(messages[currentIndex].ts) - startTimestamp;

                    cli.log.info('no messages in batch interval. top message ts: ' + Date.parse(messages[currentIndex].ts) + ' vs. current: ' + startTimestamp + " delta: " + delta);
                    return;
                }

                nitrogen.Message.sendMany(session, batch, function(err, messages) {
                    if (err) {
                        cli.log.error('failed to replay message: ' + err);
                    } else {
                        cli.log.info('replayed:');
                        console.dir(messages)
                    }
                });
            }, batchInterval);
        });
    });
};

var send = function(callback) {
    var jsonString = cli.nextArgument();
    var messageJson = JSON.parse(jsonString);

    cli.globals.startSession(function(err, session, user) {
        if (err) return callback(err);

        new nitrogen.Message(messageJson).send(session, callback);
    });
};

module.exports.execute = function(callback) {
    var command = cli.nextArgument();

    switch(command) {
        case 'help': {
            help(callback);
            break;
        }

        case 'import': {
            importFile(callback);
            break;
        }

        case 'ls': {
            list(callback);
            break;
        }

        case 'replay': {
            replay(callback);
            break;
        }

        case 'send': {
            console.log('in message execute send');
            send(callback);
            break;
        }

        default: {
            cli.log.error('unknown message command: ' + command);
            help(callback);
            break;
        }
    }
};